#include <iostream>
#include <errlog.h>
#include <astrotime.h>
#include "sub_main.h"
#include "deftypes.h"
#include "actions.h"

const std::string park_phase = "Parking: ";
const std::string init_phase = "Initialization: ";
const std::string sect_comp = "Components";
const std::string sect_gen  = "General";
const unsigned int short_delay = 3;

AstroDateTime uta;

Druid::Druid( Config& cfg ) : background( cfg, *this ), mass_enable(true) {
    settle_time  = cfg.inp<unsigned int>( sect_comp, "RCX400", "GuideSettleTime" );
    separation   = cfg.inp<double>( sect_comp, "DIMM", "Separation" );
    sep_tolrn    = cfg.inp<double>( sect_comp, "DIMM", "SepTolerance" );
    work_zone    = cfg.inp<double>( sect_comp, "RCX400", "WorkZone" );
    mass_mag_ext = cfg.inp<double>( sect_comp, "MASS", "MaxExtinct" );
    chan_fluxes  = cfg.inpv<double>( sect_comp, "MASS", "MinStarFluxes" );
    ClientSocket::AddrType dimmIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "DIMM", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "DIMM", "ThePort") );
    ClientSocket::AddrType massIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "MASS", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "MASS", "ThePort") );
    ClientSocket::AddrType tlspIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "RCX400", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "RCX400", "ThePort") );
    try {
        dimm = std::auto_ptr<ClientHandler>( new ClientHandler( dimmIP ) );
        mass = std::auto_ptr<ClientHandler>( new ClientHandler( massIP ) );
        tlsp = std::auto_ptr<ClientHandler>( new ClientHandler( tlspIP ) );
    }
    catch(const std::exception& e ) {
        ErrlogSingl::instance().write( std::string("ASM components: ") + e.what() );
        throw ErrFat( ERR_DEV_FUNC, "ASM components connection fails" );
    }
    try {
        if( dimm->wait_nonbusy( "init" ) && dimm->status( "READY" ) )
            ErrlogSingl::instance().write( init_phase + "DIMM ready" );
        if( mass->wait_nonbusy( "init" ) && mass->status( "READY" ) )
            ErrlogSingl::instance().write( init_phase + "MASS ready" );
        if( tlsp->wait_nonbusy( "init" ) && tlsp->status( "READY" ) )
            ErrlogSingl::instance().write( init_phase + "RCX400 ready" );
        if( tlsp->status( "ERFAT" ) ) throw std::runtime_error(" Server fatal error" );
    }
    catch(const std::exception& e ) {
        sleep(5*short_delay);
        if( dimm->get_status( "READY" ) ) dimm->wait_nonbusy( "park" );
        if( mass->get_status( "READY" ) ) mass->wait_nonbusy( "park" );
        if( tlsp->get_status( "READY" ) ) tlsp->wait_nonbusy( "park" );
        ErrlogSingl::instance().write( std::string("Initialization: ") + e.what() );
        throw ErrFat( ERR_DEV_FUNC, "ASM components initialization fails" );
    }
    star_fluxes = std::vector<double>(4);
}

Druid::~Druid() {
    try {
        background.run();
        sleep(short_delay);
    }
    catch(const std::exception& e ) {
        ErrlogSingl::instance().write( std::string("Finalize work: ") + e.what() );
    }
    if( park( tlsp.get() ) ) ErrlogSingl::instance().write( park_phase + "RCX400 stopped" );
    if( park( mass.get() ) ) ErrlogSingl::instance().write( park_phase + "MASS stopped" );
    if( park( dimm.get() ) ) ErrlogSingl::instance().write( park_phase + "DIMM stopped" );
//    std::cerr << "Druid  destructor" << std::endl;
}

bool Druid::park( ClientHandler* comp ) {
    try {
        if( comp->status( "BUSY" ) ) {
            comp->command( "stop" );
            sleep( 2*short_delay );
        }
        if( comp->wait_nonbusy( "park" ) && comp->status( "PARKED" ) ) return true;
    }
    catch(const std::exception& e ) {
        ErrlogSingl::instance().write( std::string("Park Druid components: ") + e.what() );
    }
    return false;
}

bool Druid::point_star( const ObjectInfo& obj ) try {
    int acount = 0;
    if( !background.no_needed() ) background.run();             // Background measurement before repointing
    ErrlogSingl::instance().write( "Start point to " + ::Name( obj ) );
    while( acount < max_retries ) {
        if( acount > 0 ) {
            ErrlogSingl::instance().write( ::Name( obj ) + " is not found. Will try again" );
            sleep( settle_time/2 );
        }
        tlsp->command( "run ra=\"" + ::RA( obj ) + "\" dec=\"" + ::DEC( obj ) +"\"" );
        if( tlsp->status( "BUSY" ) ) {
            sleep( settle_time/4 );
            continue;
        }
        if( tlsp->status( "READY" ) ) {
            ErrlogSingl::instance().write( ::Name( obj ) + " is found" );
            dimm->wait_nonbusy( "set object=\"" + OBJECT( obj ) + "\"" );
            mass->wait_nonbusy( "set object=\"" + OBJECT( obj ) + "\"" );
            background.run();                                   // Background measurement after pointing
            star = obj;
            return center_star();
        }
        acount++;
    }
    ErrlogSingl::instance().write( ::Name( obj ) + " is not found! Ask eigther star" );
    return false;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Pointing: ") + e.what() );
    return false;
}

bool Druid::center_star() try {
    int acount = 0;
    while( acount < max_retries ) {
        if( acount > 0 ) {
            ErrlogSingl::instance().write( "Star centering failed. Will try again" );
            sleep(settle_time/4);
        }
        tlsp->command( "run center" );
        if( tlsp->status( "READY" ) ) return true;
        acount++;
    }
    ErrlogSingl::instance().write( "Star centering failed" );
    return false;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Centering: ") + e.what() );
    return false;
}

bool Druid::check_star( const ObjectInfo& obj ) try {
    errflux = 1;
    if( mass_enable ) {
        mass->command( "run flux" );
        if( !check_flux() ) {
            center_star();
            mass->command( "run flux" );
            check_flux();
        }
        if( !check_mass_mag() ) return false;
    }
    start_dimm_guiding();
    dimm->command( "get offset" );                      //  
    point<double>offs = point<double>( dimm->get_value<double>( "X" ), dimm->get_value<double>( "Y" ) );
    if( offs.distance() < work_zone/5 ) {
        ErrlogSingl::instance().write( "DIMM guiding works" );
    }
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Checking: ") + e.what() );
    return false;
}

bool Druid::measurement() try {
    int max_att = max_attempts;
    dimm->nonblocked( "run" );
    while( dimm->status( "BUSY" ) ) {
        if(max_att-- < 0 ) throw std::runtime_error("DIMM very busy" );
        usleep( 500000 );
        dimm->nonblocked( "run" );
    }
    mass->nonblocked( "run" );
    dimm->wait_command();
    mass->wait_command();
    if( !mass->status( "READY" ) || !dimm->status( "READY" ) ) errcount++;
    if( !check_flux() ) return false;
    if( !check_mass_mag() ) return false;
    mass_mag_set.push_back( -1.086*log(star_fluxes[3]) );    
    return true;
}
catch(const ErrFat& e ) {
    errcount++;
    if( e.err() == ERR_SOCK_BAD ) throw ErrFat( e );
    return false;
}
catch(const std::exception& e ) {
    errcount++;
    ErrlogSingl::instance().write( std::string("Measurement: ") + e.what() );
    return false;
}

bool Druid::dimm_measurement() try {
    int max_att = 8;
    mass->command( "set hv=off" );
    dimm->nonblocked( "run" );
    while( dimm->status( "BUSY" ) ) {
        if(max_att-- < 0 ) throw std::runtime_error("DIMM very busy" );
        usleep( 500000 );
        dimm->nonblocked( "run" );
    }
    dimm->wait_command();
    if( !dimm->status( "READY" ) ) errcount++;
    return true;
}
catch(const ErrFat& e ) {
    errcount++;
    if( e.err() == ERR_SOCK_BAD ) throw ErrFat( e );
    return false;
}
catch(const std::exception& e ) {
    errcount++;
    ErrlogSingl::instance().write( std::string("DIMM Measurement: ") + e.what() );
    return false;
}

bool Druid::beginning() try {
    errcount = 0;
    mass_mag_set.clear();
    return focus_dimm();
}
catch(const ErrFat& e ) {
    if( e.err() == ERR_SOCK_BAD ) throw ErrFat( e );
    return false;
}
catch(const std::exception& e ) {
    errcount++;
    ErrlogSingl::instance().write( std::string("Measurement: ") + e.what() );
    return false;
}

bool Druid::check_status() {
    message<int,0>( "Measurement set is completed. Errors=", errcount );
    check_aperture_ratio();
    
    if( background.needed() ) check_background();
//    if( !check_mass_mag() ) return false;
    ErrlogSingl::instance().write( "Checking successful" );
    return true;
}

bool Druid::check_experiment_status( unsigned int lse ) {
    message<int,0>( "Experimental set is completed. Errors=", errcount );
    if( errcount >= lse ) return false;
    return true;
}

bool Druid::check_background() try {
    if( background.run() ) {
        tlsp->command( "run center" );
        start_dimm_guiding();
    }
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Background checking: ") + e.what() );
    return false;
}

void Druid::regime( const std::string& s ) {
    static std::string prevs = "normal";
    if( s != prevs ) {
        if( s == "normal" ) {
            camera("high");
            mass_enable = true;
        }
        if( s == "massdimm" ) {
            camera("medium");
            mass_enable = true;
        }
        if( s == "dimm" ) {
            camera("low");
            mass_enable = false;
        }
        prevs = s;
    }
}

void Druid::camera( const std::string& s ) {
    static std::string prevs = "high";
    if( s != prevs ) {
        tlsp->command( "set camera="+s );
        prevs = s;
    }
}

void Druid::focusing( double f ) {
    if( fabs( f ) < 0.002 ) return;
    f = sign(f)*min( fabs(f), 0.05 );
    focus += f;
    std::ostringstream ostr;
    ostr << "run focus=" << std::fixed << std::setprecision(3) << f;
    tlsp->command( ostr.str() );
}

double Druid::check_separation() {
    dimm->command( "run center" );
    if( dimm->status( "ERFAT" ) ) return 0.0;
    if( dimm->status( "BUSY" ) ) sleep(2);
    dimm->command( "get images" );
    if( dimm->get_value<double>( "dX" )==0 && dimm->get_value<double>( "dY" )==0 ) return 0.0;
    return separation - dimm->get_value<double>( "dX" );
}

bool Druid::focus_dimm() {
    const int ittmax = 2;
    int icount = 0;
    double dS = check_separation();
    while( fabs(dS) > sep_tolrn && icount++ < ittmax ) {
        focusing( 0.004*dS );
        dS = check_separation();
    }
    if( uta.currentSec() - guide_last > 1000.0 ) {
        guide_last = uta.currentSec();
        tlsp->command( "get center" );
        message<int,0>( "Optical center X = ", tlsp->get_value<double>( "X" ), " Y = ", tlsp->get_value<double>( "Y" ) );
    }
    return fabs(dS) < sep_tolrn;
}

bool Druid::check_flux() {
    if( mass->command( "get flux" ) ) {
        star_fluxes[0] = mass->get_value<double>( "Flux_A" ) - background.get()[0];
        star_fluxes[1] = mass->get_value<double>( "Flux_B" ) - background.get()[1];
        star_fluxes[2] = mass->get_value<double>( "Flux_C" ) - background.get()[2];
        star_fluxes[3] = mass->get_value<double>( "Flux_D" ) - background.get()[3];
    }
    for( unsigned int i = 0; i < 4; i++ ) if( star_fluxes[i] < chan_fluxes[i]) return false;
    return true;
}

bool Druid::check_aperture_ratio() {
    std::vector<double> ratio(4);
    for( unsigned int i = 0; i < 4; i++ ) ratio.at(i) = log(star_fluxes.at(i)/chan_fluxes.at(i));
    message<double,3>( "Channel ratios: ", ratio[0]-ratio[3], ratio[1]-ratio[3], ratio[2]-ratio[3] );
    if( std::abs( ratio[2]-ratio[3] ) > 0.100 ) tlsp->command( "set heater=50" );
    if( std::abs( ratio[2]-ratio[3] ) < 0.025 ) tlsp->command( "set heater=10" );
    return true;
}

bool Druid::cnt( bool wh) {
    if( wh ) {
        mass->command( "set cnt=on" );
        cnt_on = true;
    }
    if( !wh && cnt_on ) {
        mass->command( "set cnt=off" );
        cnt_on = false;
    }
    return cnt_on;
}

bool Druid::photo_stability( unsigned int nm ){
    if( errcount > nm/2 ) return false;
    if( mass_mag_set.size() > nm - 1 ) {
        double minflux = *std::min_element(mass_mag_set.begin(), mass_mag_set.end());
        double maxflux = *std::max_element(mass_mag_set.begin(), mass_mag_set.end());
        if( maxflux - minflux <  0.02 ) return true;
    }
    return false;
}

bool Druid::check_mass_mag() {
    double mass_cnt = MASScount( star );
    message<double,0>( "Calculated MASS Flux_D: ", mass_cnt, ", measured: ", star_fluxes[3] );
    double instr_mag =  -1.086*log( star_fluxes[3]/mass_cnt );
    message<double,3>( "Relative flux, mag: ",  instr_mag, " Mz= ", Mz(star) );
    if( instr_mag > mass_mag_ext ) {
        ErrlogSingl::instance().write( "Low star fluxes" );
        if( errflux++ > 3 ) return false;
    } else {
        errflux = 0;
    }
    return true;
}

bool Druid::start_guiding() {
    tlsp->wait_nonbusy( "set guide=on" );
    if( tlsp->status( "READY" ) ) { sleep(short_delay); return true; }
    ErrlogSingl::instance().write( "Guiding start failed" );
    return false;
}

bool Druid::start_dimm_guiding() {
    tlsp->wait_nonbusy( "set guide=dimm" );
    if( tlsp->status( "READY" ) ) {
        sleep( settle_time );
        guide_last = uta.currentSec();
        return true;
    }
    ErrlogSingl::instance().write( "DIMM guiding start failed" );
    return false;
}

bool Druid::stop_guiding() {
    tlsp->wait_nonbusy( "set guide=off" );
    if( tlsp->status( "READY" ) ) { sleep(short_delay); return true; }
    ErrlogSingl::instance().write( "Guiding stop failed" );
    return false;
}

const double limit_p_value = 1.20;

Background::Background( Config& cfg, Druid& ac ) : act(ac), value(-1), errcount(0) {
    displace = cfg.inp<double>( sect_gen, "Background", "Displacement" );
    period   = cfg.inp<double>( sect_gen, "Background", "Period" );
    same     = cfg.inp<double>( sect_gen, "Background", "MinimalTime" );
    maximal  = cfg.inp<double>( sect_gen, "Background", "MaxValue" );
    last = uta.currentSec();
}

bool Background::needed() {
    return ( uta.currentSec() < last + period )? false : true;
}

bool Background::no_needed() {
    return ( uta.currentSec() < last + same )? true : false;
}

bool Background::run() {
    std::vector<double> currback;
    int acount = 0;
    if( !act.mass_enable ) return false;
    do {
        currback = measure();
        if( non_poiss < limit_p_value ) {
            backgrounds = currback;
            message<double,2>( "Background in D aperture: ", backgrounds[3] );
            return true;
        }        
        ErrlogSingl::instance().write( "Star in the aperture, Will try again" );
        act.tlsp->command( "run center" );
        if( act.tlsp->status( "ERFAT" ) ) return false;
    } while( acount++ < max_count );
    ErrlogSingl::instance().write( "Background estimation problem" );
    return false;
}

std::vector<double> Background::measure() {
    std::vector<double> fluxes(4);
    std::ostringstream ostr;
    ostr << std::fixed << std::setprecision(0) << "run dy=" << displace;
    act.tlsp->command( ostr.str() );
    act.mass->command( "run background" );
    if( act.mass->status( "READY" ) ) {
        act.mass->command( "get flux" );
        fluxes[0] = act.mass->get_value<double>("Flux_A");
        fluxes[1] = act.mass->get_value<double>("Flux_B");
        fluxes[2] = act.mass->get_value<double>("Flux_C");
        fluxes[3] = act.mass->get_value<double>("Flux_D");
        act.mass->command( "get vars" );
        non_poiss = act.mass->get_value<double>("Vars_D")*fluxes[3];
        last = uta.currentSec();
        std::ostringstream o2str;
        o2str << std::fixed << std::setprecision(0) << "run dy=" << -displace;
        act.tlsp->command( o2str.str() );
    }
    return fluxes;
}
