#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <errlog.h>
#include <astrotime.h>
#include <configs.h>
#include <signal.h>

#include "deftypes.h"
#include "actions.h"
#include "defnames.h"
#include "conditions.h"
#include "object.h"

SV_PAR Condition::cmd = { std::string( program_cfg ), false, false, -1,  -1, -1, -1, -1  };

bool Condition::terminate = false;
const std::string sect_comp = "Components";

AstroDateTime Condition::ut_con;

Condition::Condition() : acfg( cmd.cfgfile ),  sun( acfg ), wind(), sky_sens(), supp() {
    meteoIP = IP4Addr( acfg.inp<std::string>( sect_comp,  "Meteo", "TheHost"),
                       acfg.inp<unsigned int>( sect_comp, "Meteo", "ThePort") );
    skyIP   = IP4Addr( acfg.inp<std::string>( sect_comp,  "SkySensor", "TheHost"),
                       acfg.inp<unsigned int>( sect_comp, "SkySensor", "ThePort") );

    civil_twil = acfg.inp<double>( "General", "Common", "CivilNight")*M_PI/180;
    astro_twil = acfg.inp<double>( "General", "Common", "AstroNight")*M_PI/180;
    if( cmd.start_time < 0 ) cmd.start_time = night_time_interval( sun.set( civil_twil )  + utc() );
    if( cmd.final_time < 0 ) cmd.final_time = night_time_interval( sun.rise( civil_twil ) + utc() );
    cmd.night_time = night_time_interval(  sun.set( astro_twil ) + utc() );
    cmd.twill_time = night_time_interval( sun.rise( astro_twil ) + utc() );
    cmd.watch_time = cmd.twill_time - acfg.inp<double>( "General", "Common", "MinimalTime");
    sens_period = acfg.inp<unsigned int>( "General", "Common", "PollingTime");
    error_pause = acfg.inp<unsigned int>( "General", "Common", "ErrPause");

    wind.high_limit = acfg.inp<double>( sect_comp, "Meteo", "WindHigh");
    wind.low_limit  = acfg.inp<double>( sect_comp, "Meteo", "WindLow");
    wind.hold_time  = acfg.inp<double>( sect_comp, "Meteo", "HoldTime");
    sky_sens.high_limit = acfg.inp<double>( sect_comp, "SkySensor", "TempHigh");
    sky_sens.low_limit  = acfg.inp<double>( sect_comp, "SkySensor", "TempLow");
    sky_sens.hold_time  = acfg.inp<double>( sect_comp, "SkySensor", "HoldTime");
    supp.high_limit = acfg.inp<double>( sect_comp, "Meteo", "PowerOff");
    supp.low_limit  = acfg.inp<double>( sect_comp, "Meteo", "PowerOn");
    supp.hold_time  = acfg.inp<double>( sect_comp, "Meteo", "HoldPower");

    meteo = std::auto_ptr<ClientHandler>( new ClientHandler( meteoIP ) );
    sky   = std::auto_ptr<ClientHandler>( new ClientHandler( skyIP ) );
//    std::cerr << "Condition constructor" << std::endl;
}

double Condition::ltc(){
    ut_con.local();
    return ut_con.hour() + (ut_con.min() + ut_con.sec()/60.0)/60.0;
}

double Condition::utc(){
    ut_con.current();
    return night_time_interval( ut_con.hour() + (ut_con.min() + ut_con.sec()/60.0)/60.0 );
}

double Condition::restart_time(){
    return cmd.start_time = utc() + error_pause/3600.0;
}

bool Condition::supp_watch() {
    return true;
}

bool Condition::wind_watch() {
    wind.errcount++;
    meteo->command( "get wind" );
    wind.value = meteo->get_value<double>( "WIND" );
    if( wind.value > wind.high_limit && wind.flag ) {
        wind.flag = false;
        ErrlogSingl::instance().write( "Strong wind is detected!" );
    }
    if( wind.value > wind.high_limit ) wind.last_event = ut_con.currentSec();
    if( wind.value <= wind.low_limit && !wind.flag && ut_con.currentSec() - wind.last_event > wind.hold_time ) {
        wind.flag = true;
        ErrlogSingl::instance().write( "Light wind" );
    }
    wind.errcount = 0;
    return  wind.flag;
}

bool Condition::sky_watch() {
    sky_sens.errcount++;
    sky->command( "get sky" );
    sky_sens.value  = sky->get_value<double>( "TEMP_SKY" );
    sky_sens.value += sky->get_value<double>( "TEMP_SEN" );
    sky_sens.value -= sky->get_value<double>( "TEMP_AMB" );
    sky_sens.value += ( sky->get_value<int>( "SKY" ) == 1 || sky->get_value<int>( "SKY" ) == 2 )? 0 : 10.0;
    if( sky_sens.value > sky_sens.high_limit && sky_sens.flag ) {
        sky_sens.flag = false;
        ErrlogSingl::instance().write( "Clouds are detected" );
    }
    if( sky_sens.value > sky_sens.high_limit ) sky_sens.last_event = ut_con.currentSec();
    if( sky_sens.value <= sky_sens.low_limit && !sky_sens.flag && ut_con.currentSec() - sky_sens.last_event > sky_sens.hold_time ) {
        sky_sens.flag = true;
        ErrlogSingl::instance().write( "Sky is clear!" );
    }
    sky_sens.errcount = 0;
    return  sky_sens.flag;
}

bool Condition::good() try { // ,   
    bool ret =  wind_watch() & supp_watch() & sky_watch();
    std::ostringstream ostr;
    ostr << " Sun altitude: " << std::fixed << std::setprecision(1) << 180*sun.sun_altitude()/M_PI;
    ErrlogSingl::instance().write( "Conditions: " + status() + ostr.str() );
    return ret;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Meteo Service: ") + e.what() );
    if( sky_sens.errcount + wind.errcount + supp.errcount > 4 ) {
        ErrlogSingl::instance().write( "Meteo Service fatal error!" );
        return false;
    }
    return sky_sens.flag & wind.flag & supp.flag;
}

bool Condition::bad() {                     // bad conditions for measurements
    return !good();
}

bool Condition::night_time() { 
    return utc() > cmd.night_time && utc() < cmd.twill_time;
}

bool Condition::evening_time() {
    return utc() > cmd.start_time && utc() < cmd.night_time;
}

bool Condition::morning_time() {
    return utc() > cmd.twill_time && utc() < cmd.final_time && sun.sun_altitude() < civil_twil;
}

std::string Condition::status() { 
    std::ostringstream ostr;
    ostr << ( wind.flag ? "Light wind: " : "Strong wind: ") << std::fixed << std::setprecision(1) << wind.value;
    ostr << ( sky_sens.flag ? " Clear sky: " : " Clouds: ") << std::fixed << std::showpos << std::setprecision(1) << sky_sens.value;
    return  ostr.str();
}

void Condition::wait_conditions() {
    timespec a = { (long)sens_period, 0 };
    while( ( utc() < cmd.start_time) | bad() ) {
        if( utc() > cmd.watch_time ) throw ErrPark( WARNING, "Night watching is pass away" );
        if( !terminate ) nanosleep( &a, NULL );
        if( terminate  ) throw ErrPark( WARNING, "External termination" );
    }
}
