#include <iostream>
#include <fstream>
#include <algorithm>
#include <functional>
#include "defnames.h"
#include "deftypes.h"
#include "object.h"

AstroDateTime ot;
//1122 Del_Per 03 42 56 +47 47 15  3.01 -0.13 B55 B5IIIe        1  var0.05 @1.2NW 4.2V
const unsigned int list_max_size = 5;

Object::Object( Config& cfg, bool m ) : astro( cfg ), current(0), mode(m) {
    ASM = new Geo( cfg.site_coords() );
    std::string cat  = cfg.inp<std::string>( "General", "Objects", "Catalogue");
    select.min_alt   = cfg.inp<double>( "General", "Objects", "MinAltitude" )*M_PI/180;
    select.moon_dist = cfg.inp<double>( "General", "Objects", "MoonZone" )*M_PI/180;
    select.hold_time = cfg.inp<double>( "General", "Objects", "HoldTime" );
    select.level     = cfg.inp<double>( "General", "Objects", "Selection" );
    phot_time = ot.currentSec() - select.hold_time/2;
    fix_time = ot.currentSec() - select.hold_time;
    // Extending photometry list
    double phot_bot = M_PI/2 - ASM->latitude() - M_PI/5;
    double phot_top = M_PI/2 - ASM->latitude() + M_PI/5;
    std::string obj;
    std::ifstream star_list( (program_dat + cat).c_str() );
    Degree dec;
    Hour ra;
    while( getline( star_list, obj ) ) {
        if( obj.size() < 1 || obj.at(0) == '#' ) continue;
        ObjectInfo t;
        std::istringstream istr( obj.substr( 0, 47 ) );
        istr >> t.HR >> t.name >> ra >> dec >> t.Vmag >> t.BVcolor >> t.spectra;
        t.coords = Coord<equ>( ra, dec, ASM );
        list.push_back( t );
        if( t.coords.dec() > phot_bot && t.coords.dec() < phot_top &&
            t.BVcolor < 0.5 ) photo.push_back( t );
        if( t.Vmag < 2.3 ) exper.push_back( t );
    }
    message< unsigned int,0 >( "Stars for photometry: ", photo.size() );
    message< unsigned int,0 >( "Stars for twilling measurements: ", exper.size() );
//    std::cerr << "Object constructor" << std::endl;
}

int Object::ask_observer( const ObjectInfo& s ) {
    if( !mode ) return 'Y';
    std::cout << OBJECT(s) << " " << std::udeg << s.coords.az() << " " << s.coords.alt()
    << " " << std::fixed << std::setprecision(2) << 1.0/sin( s.coords.alt() ) << " ?" << std::flush;
    char ans = ' ';
    do { std::cin >> ans;
        ans = toupper(ans);
    } while( ans != 'Y' && ans != 'P' && ans != 'N' );
    return ans;
}

class better {
    Constraint& criteria;
    Coord<equ> moon;
public:
    better( Constraint& sel, Coord<equ> mc ) : criteria(sel), moon(mc) {}
    bool operator()(const ObjectInfo& b1, const ObjectInfo& b2) {
        return estimation( b1 ) >  estimation( b2 );
    }
    double  estimation( const ObjectInfo& s ) {
        Coord<equ> t1 = s.coords;
        double z1 = sint(t1.ha());
        double s1 = t1.star_set( criteria.min_alt );
        return ( t1.alt() > M_PI/6 && fabs( z1 ) < s1 && t1.distance( moon ) > criteria.moon_dist ) ?
                min( 2.0, max( 4.10 - MASSmag( s ), 0.0 ) )*min( ( s1 - z1 ), criteria.hold_time*M_PI/43200 )  : 0.0;
    }
};


int Object::get_auto( ObjectInfo& curr ) {
    Coord<equ> moon = astro.mini_moon( ot.julian()+ot.currentSec()/86400.0-0.5 );
    better bb( select, moon );
    std::partial_sort( list.begin(), list.begin()+list_max_size, list.end(), bb );
    star_aviable = 0;
    for( unsigned int i = 0; i < list_max_size && bb.estimation( list[i] ) > select.level ; i++ ) {
        Coord<equ> star = list[i].coords;
        star_aviable++;
        std::ostringstream ostr;
        ostr << OBJECT(list[i]) << " " << std::udeg << star.az() << " " << star.alt()
             << std::fixed << std::setprecision(2) << " " << MASSmag( list[i] )
             << " " << bb.estimation(list[i]) << " " << star.distance(moon)*180.0/M_PI;
        ErrlogSingl::instance().write( ostr.str() );
    }
    current = 0;
    curr = list[current];
    if(bb.estimation( curr ) < select.level ) throw ErrFat( ERR_NOTHING_DONE, "No star for measurements" );
    fix_time = ot.currentSec();
    return curr.HR;
}

int Object::get_auto_exper( ObjectInfo& curr ) {
    Coord<equ> moon = astro.mini_moon( ot.julian()+ot.currentSec()/86400.0-0.5 );
    better bb( select, moon );
    std::partial_sort( exper.begin(), exper.begin()+list_max_size, exper.end(), bb );
    star_aviable_exper = 0;
    for( unsigned int i = 0; i < list_max_size && bb.estimation( exper[i] ) > select.level ; i++ ) {
        Coord<equ> star = exper[i].coords;
        star_aviable_exper++;
        std::ostringstream ostr;
        ostr << OBJECT(exper[i]) << " " << std::udeg << star.az() << " " << star.alt()
             << std::fixed << std::setprecision(2) << " " << MASSmag( exper[i] )
             << " " << bb.estimation(exper[i]) << " " << star.distance(moon)*180.0/M_PI;
        ErrlogSingl::instance().write( ostr.str() );
    }
    current = 0;   
    curr = exper[current];
    if(bb.estimation( curr ) < select.level ) throw ErrFat( ERR_NOTHING_DONE, "No star for twilling measurements" );
    fix_time = ot.currentSec();
    return curr.HR;
}

int Object::get_next( ObjectInfo& curr ) {
    current++;
    current = current%star_aviable;
    curr = list[current];
    return 0;
}

int Object::get_next_exper( ObjectInfo& curr ) {
    current++;
    current = current%star_aviable_exper;
    curr = exper[current];
    return 0;
}

int Object::get_current( ObjectInfo& curr ) {
    better cb( select, astro.mini_moon( ot.julian()+ot.currentSec()/86400.0-0.5 ) );
    curr = list[current];
    if( ot.currentSec() < fix_time + select.hold_time && cb.estimation(curr) > select.level ) {
        return curr.HR;
    } else {
        if( !mode ) return get_auto( curr );
    }
    int choice = ask_observer( curr );
    if( choice == 'Y' ) {
        fix_time = ot.currentSec();
        return curr.HR;
    }
    if( choice == 'P' ) {
        current = ( current == 0 )? list.size() - 1 : current - 1;
        return get_current( curr );
    }
    if( choice == 'N' ) {
        current = ( current == list.size()-1 )? 0 : current + 1;
        return get_current( curr );
    }
    return curr.HR;
}

int Object::get_current_exper( ObjectInfo& curr ) {
    better cb( select, astro.mini_moon( ot.julian()+ot.currentSec()/86400.0-0.5 ) );
    curr = exper[current];
    if( ot.currentSec() < fix_time + select.hold_time && cb.estimation(curr) > select.level ) {
        return curr.HR;
    } else {
        if( !mode ) return get_auto_exper( curr );
    }
    return curr.HR;
}

class coequal {
    Constraint& criteria;
    Coord<equ> moon;
    Coord<equ> star;
public:
    coequal( Constraint& sel, const Coord<equ>& mc, const Coord<equ>& cs ) : criteria(sel), moon(mc), star(cs) {}
    bool operator()(const ObjectInfo& b1, const ObjectInfo& b2) {
        return estimation( b1 ) <  estimation( b2 );
    }
    double estimation( const ObjectInfo& s ) {
        Coord<equ> t1 = s.coords;
        return ( t1.distance( moon ) > criteria.moon_dist && star.distance( t1 ) > 0.01 ) ?
                 hump( star.Mz() - t1.Mz(), 0.04 )*( 1.0 + std::abs( star.ra()-t1.ra() )/M_PI ) : 0.0;
    }
};

class closest {
    Constraint& criteria;
    Coord<equ> moon;
    Coord<equ> star;
public:
    closest( Constraint& sel, const Coord<equ>& mc, const Coord<equ>& cs ) : criteria(sel), moon(mc), star(cs) {}
    bool operator()(const ObjectInfo& b1, const ObjectInfo& b2) {
        return estimation( b1 ) <  estimation( b2 );
    }
    double estimation( const ObjectInfo& s ) {
        Coord<equ> t1 = s.coords;
        return ( t1.distance( moon ) > criteria.moon_dist ) ?
                 hump( star.Mz() - t1.Mz(), 0.1 )*hump( star.ra()-t1.ra(), M_PI/2 ) : 0.0;
    }
};

class extinct {
    Constraint& criteria;
    Coord<equ> moon;
    Coord<equ> star;
public:
    extinct( Constraint& sel, const Coord<equ>& mc, const Coord<equ>& cs ) : criteria(sel), moon(mc), star(cs) {}
    bool operator()(const ObjectInfo& b1, const ObjectInfo& b2) {
        return estimation( b1 ) <  estimation( b2 );
    }
    double estimation( const ObjectInfo& s ) {
        Coord<equ> t1 = s.coords;
        return ( t1.distance( moon ) > criteria.moon_dist && std::abs( t1.ha() ) <  M_PI/2 ) ?
                 hump( star.Mz() - t1.Mz() + 0.7, 0.2 ) : 0.0;
    }
};

int Object::get_connected( std::vector<ObjectInfo>& photo_stars ) {
//    bool nopair = false;
    if( ot.currentSec() < phot_time + 0.5*select.hold_time ) return 0;
    photo_stars.clear();
    ErrlogSingl::instance().write( "Look for photometry program" );
    Coord<equ> moon = astro.mini_moon( ot.julian()+ot.currentSec()/86400.0-0.5 );
    closest fs( select, moon, list[current].coords );
    ObjectInfo s0 = *std::max_element( photo.begin(), photo.end(), fs );
    if( fs.estimation( s0 ) < select.level ) {
        ErrlogSingl::instance().write( "No star for photo " );
        return 0;
    }
    photo_stars.push_back( s0 );
    coequal ss( select, moon, s0.coords );
    ObjectInfo s1 = *std::max_element( photo.begin(), photo.end(), ss );
    if( ss.estimation( s1 ) < select.level ) {
        ErrlogSingl::instance().write( "No pair found " );
//        nopair = true;
    } else {
        photo_stars.push_back( s1 );
    }
    extinct es( select, moon, s0.coords );
    ObjectInfo s2 = *std::max_element( photo.begin(), photo.end(), es );
    if( es.estimation( s2 ) > select.level ) {
        message< double, 2 >( "Found extinction star: ", es.estimation( s2 ) );
        photo_stars.push_back( s2 );
    } else {
        ErrlogSingl::instance().write( "No extinction star" );
        return 0;
    }
    for( unsigned int i = 0; i < photo_stars.size(); i++ ) {
        Coord<equ> star = photo_stars[i].coords;
        std::ostringstream ostr;
        ostr << OBJECT(photo_stars[i]) << " " << std::udeg << star.az() << " " << star.alt()
             << std::fixed << std::setprecision(2) << " " << MASSmag( photo_stars[i] )
             << " " << star.Mz();
        ErrlogSingl::instance().write( ostr.str() );
    }
    phot_time = ot.currentSec();
    return photo_stars[0].HR;
}

SolarSys::SolarSys( Config& cfg ) {
    AstroDateTime ut;
    ASM = new Geo( cfg.site_coords() );
    sun = mini_sun( ut.julian()+0.5 );
}

Coord<ecl> SolarSys::mini_sun( double jd ) {
    double t = ( jd - 2451545.0 )/36525.0;
    const double p2 = 2*M_PI;
    double m  = p2 * frac( 0.993133 + 99.997361 * t );
    double dl = 6893.0 * sin( m ) + 72.0*sin( 2*m );
    double l = p2 * frac( 0.7859453 + m/p2 + ( 6191.2*t + dl )/1296.e3 );
    return Coord<ecl>( l, 0, ASM );
}

Coord<ecl> SolarSys::mini_moon( double jd ) {
    double t = ( jd - 2451545.0 )/36525.0;
    const double p2 = 2*M_PI;
    const double arc = 206264.8062;
    double l0 = frac( 0.606433 + 1336.855225*t );
    double l  = p2*frac( 0.374897 + 1325.552410*t );
    double ls = p2*frac( 0.993133 +   99.997361*t );
    double d  = p2*frac( 0.827361 + 1236.853086*t );
    double f  = p2*frac( 0.259086 + 1342.227825*t );
    double dl  = 22640*sin(l) - 4586*sin(l-2*d) + 2370*sin(2*d) + 769*sin(2*l)
               - 668*sin(ls) - 412*sin(2*f) - 212*sin(2*l-2*d) - 206*sin(l+ls-2*d)
               + 192*sin(l+2*d)- 165*sin(ls-2*d) - 125*sin(d) - 110*sin(l+ls)
               + 148*sin(l-ls) - 55*sin(2*f-2*d);
    double s  = f + (dl + 412*sin(2*f) + 541*sin(ls))/arc;
    double h  = f-2*d;
    double n  = -526*sin(h) + 44*sin(l+h) - 31*sin(h-l) - 23*sin(ls+h)
              + 11*sin(h-ls) - 25*sin(f-2*l) + 21*sin(f-l);
                            // ecliptical coordinates in radians:
    double l_moon = p2*frac( l0 + dl/1296.0e3 );
    double b_moon  = ( 18520.0*sin(s) + n )/arc;
    return Coord<ecl>( l_moon, b_moon, ASM );
}

double SolarSys::set( double h ) {
    return 12*( sun.star_set( h ) - sun.ha())/M_PI;
}

double SolarSys::rise( double h ) {
    return 12*( -sun.star_set( h ) - sun.ha())/M_PI + 24.0;
}
