#include <iostream>
#include <errlog.h>
#include <astrotime.h>

#include "deftypes.h"
#include "omicron.h"

const std::string start_phase = "Power on: ";
const std::string sect_comp = "Components";
const unsigned int short_delay = 3;

AstroDateTime ut;

Omicron::Omicron( Config& cfg, bool m ) try : man_mode(m) {
    delay_to_druid = cfg.inp<unsigned int>( sect_comp, "Dome", "StartDelay");
    delay_from_druid = cfg.inp<unsigned int>( sect_comp, "Dome", "HaltDelay");
    ClientSocket::AddrType domeIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "Dome", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "Dome", "ThePort") );
    dome = std::auto_ptr<ClientHandler>( new ClientHandler( domeIP ) );
    ClientSocket::AddrType irIP   = IP4Addr( cfg.inp<std::string>( sect_comp,  "IRSky", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "IRSky", "ThePort") );
    irsens = std::auto_ptr<ClientHandler>( new ClientHandler( irIP ) );
    shutIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "Druid", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "Druid", "ThePort") );
    host_checked = cfg.inp<std::string>( sect_comp, "Druid", "TheHost");
    
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("ASM start: ") + e.what() );
    throw ErrFat( ERR_DEV_FUNC, "ASM connection fails" );
}

Omicron::~Omicron() {
    finish();
//     char ans = 'Y';
//     if( man_mode ) {    //  !!!
//         std::cout << " ***** Power off instruments and close dome? " << std::flush;
//         std::cin >> ans;
//     }
//     if( toupper(ans) == 'Y' ) finish();
//    std::cerr << "Omicron destructor" << std::endl;
}

bool Omicron::testDruid() {
    std::string req = "ping -c1 " + host_checked + " > /dev/null";
    return system( req.c_str() ) == 0;
}

bool Omicron::haltDruid() try {
            // While empty
    return false;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("ASM finish: ") + e.what() );
    return false;
}

bool Omicron::start() try {
    irsens->command( "init" );
    if( irsens->status( "READY" ) ) ErrlogSingl::instance().write( "IR sensor starts" );
                                //        !!!
    while( !dome->get_status( "READY" ) ) {
        if( dome->status( "ERFAT" ) ) throw ErrFat( ERR_DEV_FUNC, "Dome component fails" );
        sleep( 4*short_delay );
    }
    ErrlogSingl::instance().write( "Dome component is ready" );
    dome->command( "get dome" );
    if( dome->get_value<std::string>( "dome" ) != "OPENED" ) {
        dome->command( "get position" );        
        dome->command( "run dome=open" );
        dome->command( "get dome" );
    }
    if( dome->get_value<std::string>( "dome" ) == "OPENED" ) ErrlogSingl::instance().write( "Dome is opened" );
    else throw ErrFat( ERR_DEV_FUNC, "Dome opening fails" );
    if( !testDruid() ) throw ErrFat( ERR_DEV_FUNC, "Druid start fails" );
    ErrlogSingl::instance().write( "Druid is started" );
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("ASM start: ") + e.what() );
    dome->command( "run dome=close" );
    throw ErrFat( ERR_DEV_FUNC, "ASM start fails" );
    return false;
}

bool Omicron::finish() try {
    ErrlogSingl::instance().write( "Finish of the observation run" );
    dome->command( "run dome=close" );
    dome->command( "get dome" );
    if( dome->get_value<std::string>( "dome" ) != "CLOSED" ) {
        sleep( short_delay );
        dome->command( "get position" );        
        dome->command( "run dome=close" );
        dome->command( "get dome" );
    }  
    irsens->command( "park" );
    if( irsens->status( "PARKED" ) ) ErrlogSingl::instance().write( "IR sensor stopped" );
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("ASM finish: ") + e.what() );
    return false;
}




