#include <sstream>
#include <iostream>
#include <functional>
#include <errlog.h>
#include <stdexcept>
#include "server.h"


Server::Server( const ServerSocket::AddrType& addr, const CALLF_H func ) {
    _quit = false;
    comm_handl = func;
    cbtable_socket.onAccepted = &Server::onSocketAccepted;
    cbtable_socket.onStateChanged = &Server::onSocketStateChanged;
    cbtable_socket.onError = &Server::onSocketError;
    m_socket = new ServerSocket( &cbtable_socket, this );
    m_socket->listen( addr );
}

Server::~Server(){
    while(m_clients.size()){
        delete *(m_clients.begin());
        m_clients.remove(*(m_clients.begin()));
    }
    delete m_socket;
}

void Server::listen(const ServerSocket::AddrType& addr){
    m_socket->listen(addr);
}

void Server::close(){
    m_socket->close();
}

void Server::onSocketStateChanged( const SocketBase::SocketState state ){
}

void Server::removeDisconnectedClients() {
    while(m_clientsToRemove.size()) {
        m_clients.remove( m_clientsToRemove.top() );
        delete m_clientsToRemove.top();
        m_clientsToRemove.pop();
    }
}

void Server::onSocketAccepted(){
    Client* c = new Client( this );
    m_clients.push_back(c);
    ClientSocket::AddrType peer_addr = c->getSocket()->getPeerName();
    std::ostringstream ws;
    ws << "Client " <<  peer_addr << " is attached";
    c->storeAddr( peer_addr );
    ErrlogSingl::instance().warning( ws.str() );
}

void Server::onSocketError( const SocketError& e ) {
    ErrlogSingl::instance().wthrow( ERR_STD_EXCP, "Server::onSocketError()");
}

void Server::onClientClosed( Client* c ) {
    m_clientsToRemove.push( c );
}

void Server::onClientRequest( Client* c ){

}

Client::Client( Server* serv ) : m_parent( serv ) {
    cbtable_socket.onDataRecieved = &Client::onSocketDataRecieved;
    cbtable_socket.onStateChanged = &Client::onSocketStateChanged;
    cbtable_socket.onError = &Client::onSocketError;
    m_socket = new ClientSocket( m_parent->socket(), &cbtable_socket, this );
}

const unsigned int connect_timeout = 15;

Client::Client( ClientSocket::AddrType& addr, ClientHandler* h ) : m_addr( addr ), handclass( h ) {
    m_parent = 0;
    cbtable_socket.onDataRecieved = &Client::onSocketDataRecieved;
    cbtable_socket.onStateChanged = &Client::onSocketStateChanged;
    cbtable_socket.onError = &Client::onSocketError;
    m_socket = new ClientSocket( &cbtable_socket, this );
    m_socket->connect(m_addr);
    sem_init( &client_sem, 0, 0 );
    waitReply( connect_timeout );
//    std::cerr << "Client constructor" << std::endl;
}

Client::~Client(){
//    std::cerr << "Client destructor" << std::endl;
    delete m_socket;
    sem_destroy( &client_sem );
}

void Client::sendRequest( const std::string&  r ) {
    sendResponse( r );
}

const unsigned int first_timeout = 10;

void Client::sendRequest( const std::string&  r, CALLF_C func ) {
    int sem_cnt;
    comm_handl = func;
    while( sem_getvalue( &this->client_sem, &sem_cnt ) == 0 && sem_cnt > 0 ) sem_wait( &this->client_sem );
    sendResponse( r );
    waitReply( first_timeout );
}

void Client::waitReply( unsigned int timeout ) {
    struct timespec locktime;
    struct timeval  firstime;
    gettimeofday( &firstime, NULL );
    locktime.tv_sec = firstime.tv_sec + timeout;
    locktime.tv_nsec = 1000*firstime.tv_usec;
    if( sem_timedwait( &this->client_sem, &locktime ) < 0 ) {
        if(  errno == ETIMEDOUT ) throw std::runtime_error("Server reply timeout" );
        if(  errno == EINTR ) {
            ErrlogSingl::instance().warning( "Signal while waiting server reply" );
            if( sem_timedwait( &this->client_sem, &locktime ) < 0 ) throw std::runtime_error("Server reply timeout" );
        } else throw std::runtime_error("Server reply unknown error" );
    }
}

void Client::sendResponse(const std::string& r) try {
    std::string tmp = r + "\n";
    (*m_socket) << tmp;
}
catch( const std::exception& e ) {
    std::ostringstream ws;
    ws << m_addr << " is disconnected!";
    throw std::runtime_error( ws.str() );
}

void Client::close(){
    m_socket->close();
}

void Client::onSocketStateChanged(const SocketBase::SocketState state) {
    std::ostringstream ws;
    if( state == SocketBase::Idle && m_socket->getState() == SocketBase::Connected ) {
        if( m_parent != 0 ) {
            m_parent->onClientClosed( this );
            ws << "Client " << m_addr << " is detached";
        } else {
            ws << "Connection to " << m_addr << " closed";
        }
        ErrlogSingl::instance().warning( ws.str() );
    }
    if( state == SocketBase::Connected ) {
        ws << "Connection to " << m_addr << " done";
        ErrlogSingl::instance().warning( ws.str() );
        sem_post( &this->client_sem );
    }
}

bool Client::isConnected() {
    while( m_socket->getState() == SocketBase::Connecting ) usleep(10000);
    return m_socket->getState() == SocketBase::Connected;
}

bool Client::messExtract( std::string& data ) {
    buff_mesgs += data;
    std::string::size_type pos = 0;
    if( ( pos = buff_mesgs.find("\n") ) != std::string::npos ){
        data = buff_mesgs.substr( 0, pos );
        buff_mesgs.erase(0,++pos);
        return true;
    } 
    data.clear();
    return false;
}

void Client::onSocketDataRecieved() {
    std::string buf;
    m_socket->getData( buf );
    while( this->messExtract( buf ) ) {
        std::string::size_type pos = 0;
        if( ( pos = buf.find( "\r" ) ) != std::string::npos ) buf.erase( pos );
        if( m_parent != 0 ) {
            this->sendResponse( (*m_parent->comm_handl)( this, buf ) );
        } else {
            (handclass->*comm_handl)( buf );
            sem_post( &this->client_sem );
        }
        buf.clear();
    }
}

void Client::onSocketError( const SocketError& e ) {
    std::ostringstream ws;
    if( e.code() == ECONNREFUSED ) {
        ws << "Connection to " << m_addr << " refused";
        sem_post( &this->client_sem );
        throw ErrFat(  (ErrorsCode)e.code(), ws.str() );
    }
    throw ErrFat(  (ErrorsCode)e.code(), e.what() );
}

